@@ -15,6 +15,9 @@ import json
 # Templates - use these to apply variables to URIs
 from jinja2 import Environment, BaseLoader
 
+# System calls for crashing
+import sys
+
 
 # Begin Supporting Classes
 class Settings:
@@ -27,6 +30,10 @@ class Settings:
                 "certificate": "",
                 "key": "",
             },
+            "headers": {
+                "content-type": "application/json",
+                "X-Allow-Overwrite": "true",
+            },
             "tls": {"validation": False},
             "verbosity": 1,
             "endpoint": "https://nsx.lab.engyak.net",
@@ -104,6 +111,7 @@ class Reliquary:
     cogitation_password = ""
     cogitation_authkey = ""
     cogitation_authcert = ""
+    cogitation_headers = ""
     cogitation_endpoint = ""
     cogitation_bibliotheca = {}
     cogitation_errors = {}
@@ -135,6 +143,7 @@ class Reliquary:
             ]
             self.cogitation_authkey = json_settings["settings"]["authentication"]["key"]
             # Load Known API Actions
+            self.cogitation_headers = json_settings["settings"]["headers"]
             self.cogitation_bibliotheca = json_settings["plays"]
             # Load HTTP Errors
             self.cogitation_errors = json_settings["errors"]
@@ -148,26 +157,33 @@ class Reliquary:
     # Functions
 
     # Do API DELETE, using basic credentials
-    def do_api_delete(self, do_api_uri):
+    def do_api_delete(self, do_api_uri, do_api_dryrun=False):
         # Perform API Processing - conditional basic authentication
         try:
-            do_api_delete_headers = {
-                "content-type": "application/json",
-                "X-Allow-Overwrite": "true",
-            }
             do_api_delete_url = self.cogitation_endpoint + do_api_uri
             self.validate_url(do_api_delete_url)
-            do_api_delete_r = requests.delete(
-                do_api_delete_url,
-                headers=do_api_delete_headers,
-                verify=self.cogitation_certvalidation,
-                auth=(self.cogitation_username, self.cogitation_password),
-            )
-            # We'll be discarding the actual `Response` object after this, but we do want to get HTTP status for erro handling
-            response_code = do_api_delete_r.status_code
-            print(response_code)
-            do_api_delete_r.raise_for_status()  # trigger an exception before trying to convert or read data. This should allow us to get good error info
-            return do_api_delete_r.text  # if HTTP status is good, save response
+            if not do_api_dryrun:
+                do_api_delete_r = requests.delete(
+                    do_api_delete_url,
+                    headers=self.cogitation_headers,
+                    verify=self.cogitation_certvalidation,
+                    auth=(self.cogitation_username, self.cogitation_password),
+                )
+                # We'll be discarding the actual `Response` object after this, but we do want to get HTTP status for erro handling
+                response_code = do_api_delete_r.status_code
+                print(response_code)
+                do_api_delete_r.raise_for_status()  # trigger an exception before trying to convert or read data. This should allow us to get good error info
+                return do_api_delete_r.text  # if HTTP status is good, save response
+            else:
+                print(
+                    json.dumps(
+                        {
+                            "do_api_get_headers": self.cogitation_headers,
+                            "do_api_get_url": do_api_delete_url,
+                        },
+                        indent=4,
+                    )
+                )
         except requests.Timeout:
             print("E1000: API Connection timeout!")
         except requests.ConnectionError as connection_error:
@@ -193,25 +209,32 @@ class Reliquary:
             exit()
 
     # Do API GET, using basic credentials
-    def do_api_get(self, do_api_uri):
+    def do_api_get(self, do_api_uri, do_api_dryrun=False):
         # Perform API Processing - conditional basic authentication
         try:
-            do_api_get_headers = {
-                "content-type": "application/json",
-                "X-Allow-Overwrite": "true",
-            }
             do_api_get_url = self.cogitation_endpoint + do_api_uri
             self.validate_url(do_api_get_url)
-            do_api_get_r = requests.get(
-                do_api_get_url,
-                headers=do_api_get_headers,
-                verify=self.cogitation_certvalidation,
-                auth=(self.cogitation_username, self.cogitation_password),
-            )
-            # We'll be discarding the actual `Response` object after this, but we do want to get HTTP status for erro handling
-            response_code = do_api_get_r.status_code
-            do_api_get_r.raise_for_status()  # trigger an exception before trying to convert or read data. This should allow us to get good error info
-            return do_api_get_r.text  # if HTTP status is good, save response
+            if not do_api_dryrun:
+                do_api_get_r = requests.get(
+                    do_api_get_url,
+                    headers=self.cogitation_headers,
+                    verify=self.cogitation_certvalidation,
+                    auth=(self.cogitation_username, self.cogitation_password),
+                )
+                # We'll be discarding the actual `Response` object after this, but we do want to get HTTP status for erro handling
+                response_code = do_api_get_r.status_code
+                do_api_get_r.raise_for_status()  # trigger an exception before trying to convert or read data. This should allow us to get good error info
+                return do_api_get_r.text  # if HTTP status is good, save response
+            else:
+                print(
+                    json.dumps(
+                        {
+                            "do_api_get_headers": self.cogitation_headers,
+                            "do_api_get_url": do_api_get_url,
+                        },
+                        indent=4,
+                    )
+                )
         except requests.Timeout:
             print("E1000: API Connection timeout!")
         except requests.ConnectionError as connection_error:
@@ -236,77 +259,129 @@ class Reliquary:
             print("E1002: Unhandled Requests exception! " + str(e))
             exit()
 
+    # Do API POST, using basic credentials
+    def do_api_pass(
+        self, do_api_uri, do_api_verb="GET", do_api_payload=False, do_api_dryrun=False
+    ):
+        # Perform API Processing - conditional basic authentication
+        try:
+            do_api_url = self.cogitation_endpoint + do_api_uri
+            self.validate_url(do_api_url)
+            if not do_api_dryrun:
+                do_api_r = requests.request(
+                    do_api_verb,
+                    url=do_api_url,
+                    headers=self.cogitation_headers,
+                    verify=self.cogitation_certvalidation,
+                    auth=(self.cogitation_username, self.cogitation_password),
+                    json=do_api_payload,
+                )
+                # We'll be discarding the actual `Response` object after this, but we do want to get HTTP status for erro handling
+                response_code = do_api_r.status_code
+                do_api_r.raise_for_status()  # trigger an exception before trying to convert or read data. This should allow us to get good error info
+                return do_api_r.text  # if HTTP status is good, save response
+            else:
+                print(
+                    json.dumps(
+                        {
+                            "do_api_get_headers": self.cogitation_headers,
+                            "do_api_get_url": do_api_url,
+                            "do_api_verb": do_api_verb,
+                            "do_api_payload": do_api_payload,
+                        },
+                        indent=4,
+                    )
+                )
+        except requests.Timeout:
+            print("E1000: API Connection timeout!")
+        except requests.ConnectionError as connection_error:
+            print(connection_error)
+        except requests.HTTPError:
+            if self.get_http_error_code(response_code):
+                print(
+                    "EA"
+                    + str(response_code)
+                    + ": HTTP Status Error "
+                    + str(response_code)
+                    + " "
+                    + self.get_http_error_code(response_code)
+                )
+                return do_api_r.text
+            else:
+                print("EA999: Unhandled HTTP Error " + str(response_code) + "!")
+                exit()  # interpet the error, then close out so we don't have to put all the rest of our code in an except statement
+        except requests.RequestException as requests_exception:
+            print(requests_exception)
+        except Exception as e:
+            print("E1002: Unhandled Requests exception! " + str(e))
+            exit()
+
     # Executor of the API calls.
     # Takes optional arguments (variables, payload) depending on which method is used
-    def namshub(self, namshub_string, namshub_variables=False, namshub_payload=False):
+    def namshub(self, namshub_string, namshub_variables=False, namshub_dryrun=False):
         # Sanitize the verb used to uppercase, fewer changes for mixup
         namshub_verb = self.get_play_verb(namshub_string).lower().upper()
         # URI is in JSON file
         namshub_resource = self.get_play_uri(namshub_string)
         # Grab the namshub payload from the json file, if it exists
-
+        namshub_payload = self.get_play_payload(namshub_string)
         # Test to see if either variables or payloads are required
         # If they're required and not present, don't proceed
         if self.get_play_requiresvariables(namshub_string) and not namshub_variables:
-            exit(
+            sys.exit(
                 "Error: Variables required by play, but not provided! Specify as a JSON dictionary with `--vars`"
             )
         if self.get_play_requiresbody(namshub_string) and not namshub_payload:
-            exit(
+            sys.exit(
                 "Error: Payload required by play, but not provided! Specify as a JSON dictionary with `-b`"
             )
         # If a payload is required and present, load it from the file before proceeding
-        elif namshub_payload:
-            namshub_payload = self.get_json_file(namshub_payload)
+        if namshub_payload:
+            namshub_payload = json.dumps(self.get_json_file_or_string(namshub_payload))
 
-        # Simplest first. Let's try the edition that doesn't need templating first
-        if not namshub_variables:
+        # Grab the variables, if it exists
+        # From there, apply templates to whatever we can.
+        if namshub_variables:
+            namshub_variables = self.get_json_file_or_string(namshub_variables)
+            namshub_resource = self.apply_template(namshub_resource, namshub_variables)
             if namshub_payload:
-                if namshub_verb == "POST":
-                    print(self.do_api_post(namshub_resource, namshub_payload))
-                elif namshub_verb == "PATCH":
-                    print(
-                        self.do_api_patch(
-                            self.get_play_uri(namshub_resource), namshub_variables
-                        )
-                    )
-            else:
-                if namshub_verb == "DELETE":
-                    print(self.do_api_delete(namshub_resource))
-                elif namshub_verb == "GET":
-                    print(self.do_api_get(namshub_resource))
+                namshub_payload = json.loads(
+                    self.apply_template(namshub_payload, namshub_variables)
+                )
 
-        # This edition will leverage J2 templating to translate any URI variables
-        elif namshub_variables:
-            if namshub_payload:
-                if namshub_verb == "POST":
-                    print(
-                        self.do_api_post(
-                            self.apply_template(
-                                self.get_play_uri(namshub_resource), namshub_variables
-                            )
-                        )
-                    )
-                elif namshub_verb == "PATCH":
-                    print(
-                        self.do_api_patch(
-                            self.get_play_uri(namshub_resource), namshub_variables
-                        )
-                    )
-            else:
-                if namshub_verb == "DELETE":
-                    print(
-                        self.do_api_delete(
-                            self.apply_template(namshub_resource, namshub_variables)
-                        )
-                    )
-                elif namshub_verb == "GET":
-                    print(self.do_api_get(namshub_resource, namshub_variables))
+        # Now that the transforms, testing, pre-processing are done, let's send to an API!
+        if namshub_verb == "GET":
+            print(self.do_api_get(namshub_resource, do_api_dryrun=namshub_dryrun))
+        elif namshub_verb == "POST" or namshub_verb == "PATCH" or namshub_verb == "PUT":
+            print(
+                self.do_api_pass(
+                    namshub_resource,
+                    do_api_payload=namshub_payload,
+                    do_api_verb=namshub_verb,
+                    do_api_dryrun=namshub_dryrun,
+                )
+            )
+        elif namshub_verb == "DELETE":
+            print(self.do_api_delete(namshub_resource))
+        else:
+            sys.exit("Unsupported API verb " + namshub_verb + "!")
 
     # One-shot to apply templates to a given object string
     def apply_template(self, apply_template_template, apply_template_variables):
-        j2template = Environment(loader=BaseLoader).from_string(apply_template_template)
-        return j2template.render(apply_template_variables)
+        try:
+            j2template = Environment(loader=BaseLoader).from_string(
+                apply_template_template
+            )
+            return j2template.render(apply_template_variables)
+        except Exception as e:
+            sys.exit(
+                "Exception applying "
+                + json.dumps(apply_template_variables)
+                + " to "
+                + apply_template_template
+                + ". Error: "
+                + str(e)
+            )
 
     def validate_url(self, validate_url_url):
         # 99% solution here is to validate that a URL provided is valid. This doesn't test it or anything
@@ -334,7 +409,12 @@ class Reliquary:
         try:
             return self.cogitation_bibliotheca[get_play_name]["uri"]
         except Exception as e:
-            exit("Exception fetching play URI: " + str(get_play_name) + str(e))
+            exit(
+                "Exception fetching play URI: "
+                + str(get_play_name)
+                + ". Error: "
+                + str(e)
+            )
 
     # Return JSON string from `dict` payload for a given api call if successful
     # If not, return false and don't crash if a KeyError is returned
@@ -383,3 +463,18 @@ class Reliquary:
         except Exception as e:
             print("E0000: Error Loading Settings File " + filename + ": " + str(e))
             exit()
+
+    def get_json_file_or_string(self, input):
+        try:
+            return json.loads(input)
+        except Exception as e:
+            if self.cogitation_verbosity:
+                print(str(e))
+            try:
+                return self.get_json_file(input)
+            except Exception as e:
+                if self.cogitation.verbosity:
+                    print(str(e))
+                sys.exit(
+                    "Error: " + input + " is neither a valid JSON string or filename!"
+                )
